from typing import Iterable

import numpy as np
import torch

from cca_zoo.deepmodels import objectives
from cca_zoo.deepmodels.dcca import DCCA
from cca_zoo.models import MCCA


class DCCA_NOI(DCCA):
    """
    A class used to fit a DCCA model by non-linear orthogonal iterations

    """

    def __init__(self, latent_dims: int, N: int,
                 encoders=None,
                 r: float = 0, rho: float = 0.2, eps: float = 1e-3, shared_target: bool = False):
        """
        Constructor class for DCCA

        :param latent_dims: # latent dimensions
        :param N: # samples used to estimate covariance
        :param encoders: list of encoder networks
        :param r: regularisation parameter of tracenorm CCA like ridge CCA
        :param rho: covariance memory like DCCA non-linear orthogonal iterations paper
        :param eps: epsilon used throughout
        :param shared_target: not used
        """
        super().__init__(latent_dims=latent_dims, encoders=encoders,
                         r=r, eps=eps)
        self.N = N
        self.covs = None
        if rho < 0 or rho > 1:
            raise ValueError(f"rho should be between 0 and 1. rho={rho}")
        self.eps = eps
        self.rho = rho
        self.shared_target = shared_target
        self.mse = torch.nn.MSELoss()
        self.linear_layers = torch.nn.ModuleList(
            [torch.nn.Linear(latent_dims, latent_dims, bias=False) for _ in range(len(encoders))])
        self.rand = torch.rand(N, self.latent_dims)

    def forward(self, *args):
        z = self.encode(*args)
        return z

    def encode(self, *args):
        z = []
        for i, (encoder, linear_layer) in enumerate(zip(self.encoders, self.linear_layers)):
            z.append(linear_layer(encoder(args[i])))
        return tuple(z)

    def loss(self, *args):
        z = self(*args)
        z_copy = [z_.detach().clone() for z_ in z]
        self.update_covariances(*z_copy)
        covariance_inv = [torch.linalg.inv(objectives.MatrixSquareRoot.apply(cov)) for
                          cov in self.covs]
        preds = [z_ @ covariance_inv[i] for i, z_ in enumerate(z_copy)]
        loss = self.mse(z[0], preds[1]) + self.mse(z[1], preds[0])
        return loss

    def update_covariances(self, *z):
        b = z[0].shape[0]
        batch_covs = [self.N * z_i.T @ z_i / b for i, z_i in enumerate(z)]
        if self.covs is not None:
            self.covs = [self.rho * self.covs[i] + (1 - self.rho) * batch_cov for i, batch_cov in enumerate(batch_covs)]
        else:
            self.covs = batch_covs

    def post_transform(self, *z_list, train=False) -> Iterable[np.ndarray]:
        if train:
            self.cca = MCCA(latent_dims=self.latent_dims)
            self.cca.fit(*z_list)
            z_list = self.cca.transform(*z_list)
        else:
            z_list = self.cca.transform(*z_list)
        return z_list



import itertools

import numpy as np
import scipy.sparse as sp
from sklearn.utils.validation import check_random_state

from cca_zoo.models import CCA, PLS, CCA_ALS, SCCA, PMD, ElasticCCA, rCCA, KCCA, KTCCA, MCCA, GCCA, TCCA, SCCA_ADMM, \
    SpanCCA, SWCCA, PLS_ALS, KGCCA

rng = check_random_state(0)
X = rng.rand(500, 20)
Y = rng.rand(500, 21)
Z = rng.rand(500, 22)
X_sp = sp.random(500, 20, density=0.5, random_state=rng)
Y_sp = sp.random(500, 21, density=0.5, random_state=rng)


def test_unregularized_methods():
    # Tests unregularized CCA methods. The idea is that all of these should give the same result.
    latent_dims = 2
    cca = CCA(latent_dims=latent_dims).fit(X, Y)
    iter = CCA_ALS(latent_dims=latent_dims, tol=1e-9, random_state=rng, stochastic=False).fit(X, Y)
    iter_pls = PLS_ALS(latent_dims=latent_dims, tol=1e-9, initialization='unregularized',
                       centre=False).fit(X, Y)
    gcca = GCCA(latent_dims=latent_dims).fit(X, Y)
    mcca = MCCA(latent_dims=latent_dims, eps=1e-9).fit(X, Y)
    kcca = KCCA(latent_dims=latent_dims).fit(X, Y)
    kgcca = KGCCA(latent_dims=latent_dims).fit(X, Y)
    tcca = TCCA(latent_dims=latent_dims).fit(X, Y)
    corr_cca = cca.score(X, Y)
    corr_iter = iter.score(X, Y)
    corr_gcca = gcca.score(X, Y)
    corr_mcca = mcca.score(X, Y)
    corr_kcca = kcca.score(X, Y)
    corr_kgcca = kgcca.score(X, Y)
    corr_tcca = kcca.score(X, Y)
    # Check the correlations from each unregularized method are the same
    assert np.testing.assert_array_almost_equal(corr_cca, corr_iter, decimal=2) is None
    assert np.testing.assert_array_almost_equal(corr_cca, corr_mcca, decimal=2) is None
    assert np.testing.assert_array_almost_equal(corr_cca, corr_gcca, decimal=2) is None
    assert np.testing.assert_array_almost_equal(corr_cca, corr_kcca, decimal=2) is None
    assert np.testing.assert_array_almost_equal(corr_cca, corr_tcca, decimal=2) is None
    assert np.testing.assert_array_almost_equal(corr_kgcca, corr_gcca, decimal=2) is None
    # Check standardized models have standard outputs
    assert np.testing.assert_allclose(np.linalg.norm(iter.transform(X, Y)[0], axis=0) ** 2, 500, rtol=0.1) is None
    assert np.testing.assert_allclose(np.linalg.norm(cca.transform(X, Y)[0], axis=0) ** 2, 500, rtol=0.1) is None
    assert np.testing.assert_allclose(np.linalg.norm(mcca.transform(X, Y)[0], axis=0) ** 2, 500, rtol=0.1) is None
    assert np.testing.assert_allclose(np.linalg.norm(kcca.transform(X, Y)[0], axis=0) ** 2, 500, rtol=0.1) is None
    assert np.testing.assert_allclose(np.linalg.norm(iter.transform(X, Y)[1], axis=0) ** 2, 500, rtol=0.1) is None
    assert np.testing.assert_allclose(np.linalg.norm(cca.transform(X, Y)[1], axis=0) ** 2, 500, rtol=0.1) is None
    assert np.testing.assert_allclose(np.linalg.norm(mcca.transform(X, Y)[1], axis=0) ** 2, 500, rtol=0.1) is None
    assert np.testing.assert_allclose(np.linalg.norm(kcca.transform(X, Y)[1], axis=0) ** 2, 500, rtol=0.1) is None


def test_sparse_input():
    # Tests unregularized CCA methods. The idea is that all of these should give the same result.
    latent_dims = 2
    cca = CCA(latent_dims=latent_dims, centre=False).fit(X_sp, Y_sp)
    iter = CCA_ALS(latent_dims=latent_dims, tol=1e-9, stochastic=False, centre=False).fit(X_sp, Y_sp)
    iter_pls = PLS_ALS(latent_dims=latent_dims, tol=1e-9, initialization='unregularized',
                       centre=False).fit(X_sp, Y_sp)
    gcca = GCCA(latent_dims=latent_dims, centre=False).fit(X_sp, Y_sp)
    mcca = MCCA(latent_dims=latent_dims, centre=False).fit(X_sp, Y_sp)
    kcca = KCCA(latent_dims=latent_dims, centre=False).fit(X_sp, Y_sp)
    scca = SCCA(latent_dims=latent_dims, centre=False, c=0.001).fit(X_sp, Y_sp)
    corr_cca = cca.score(X, Y)
    corr_iter = iter.score(X, Y)
    corr_gcca = gcca.score(X, Y)
    corr_mcca = mcca.score(X, Y)
    corr_kcca = kcca.score(X, Y)
    # Check the correlations from each unregularized method are the same
    assert np.testing.assert_array_almost_equal(corr_cca, corr_iter, decimal=2) is None
    assert np.testing.assert_array_almost_equal(corr_iter, corr_mcca, decimal=2) is None
    assert np.testing.assert_array_almost_equal(corr_iter, corr_gcca, decimal=2) is None
    assert np.testing.assert_array_almost_equal(corr_iter, corr_kcca, decimal=2) is None


def test_unregularized_multi():
    # Tests unregularized CCA methods for more than 2 views. The idea is that all of these should give the same result.
    latent_dims = 2
    cca = rCCA(latent_dims=latent_dims).fit(X, Y, Z)
    iter = CCA_ALS(latent_dims=latent_dims, stochastic=False, tol=1e-12).fit(X, Y, Z)
    gcca = GCCA(latent_dims=latent_dims).fit(X, Y, Z)
    mcca = MCCA(latent_dims=latent_dims).fit(X, Y, Z)
    kcca = KCCA(latent_dims=latent_dims).fit(X, Y, Z)
    corr_cca = cca.score(X, Y, Z)
    corr_iter = iter.score(X, Y, Z)
    corr_gcca = gcca.score(X, Y, Z)
    corr_mcca = mcca.score(X, Y, Z)
    corr_kcca = kcca.score(X, Y, Z)
    # Check the correlations from each unregularized method are the same
    assert np.testing.assert_array_almost_equal(corr_cca, corr_iter, decimal=1) is None
    assert np.testing.assert_array_almost_equal(corr_cca, corr_mcca, decimal=2) is None
    assert np.testing.assert_array_almost_equal(corr_cca, corr_gcca, decimal=2) is None
    assert np.testing.assert_array_almost_equal(corr_cca, corr_kcca, decimal=2) is None


def test_regularized_methods():
    # Test that linear regularized methods match PLS solution when using maximum regularisation.
    latent_dims = 2
    c = 1
    kernel = KCCA(latent_dims=latent_dims, c=[c, c], kernel=['linear', 'linear']).fit(X, Y)
    pls = PLS(latent_dims=latent_dims).fit(X, Y)
    gcca = GCCA(latent_dims=latent_dims, c=[c, c]).fit(X, Y)
    mcca = MCCA(latent_dims=latent_dims, c=[c, c]).fit(X, Y)
    rcca = rCCA(latent_dims=latent_dims, c=[c, c]).fit(X, Y)
    corr_gcca = gcca.score(X, Y)
    corr_mcca = mcca.score(X, Y)
    corr_kernel = kernel.score(X, Y)
    corr_pls = pls.score(X, Y)
    corr_rcca = rcca.score(X, Y)
    # Check the correlations from each unregularized method are the same
    # assert np.testing.assert_array_almost_equal(corr_pls, corr_gcca, decimal=2))
    assert np.testing.assert_array_almost_equal(corr_pls, corr_mcca, decimal=1) is None
    assert np.testing.assert_array_almost_equal(corr_pls, corr_kernel, decimal=1) is None
    assert np.testing.assert_array_almost_equal(corr_pls, corr_rcca, decimal=1) is None


def test_non_negative_methods():
    latent_dims = 2
    nnelasticca = ElasticCCA(latent_dims=latent_dims, tol=1e-9, positive=True, l1_ratio=[0.5, 0.5],
                             c=[1e-4, 1e-5]).fit(X, Y)
    als = CCA_ALS(latent_dims=latent_dims, tol=1e-9).fit(X, Y)
    nnals = CCA_ALS(latent_dims=latent_dims, tol=1e-9, positive=True).fit(X, Y)
    nnscca = SCCA(latent_dims=latent_dims, tol=1e-9, positive=True, c=[1e-4, 1e-5]).fit(X, Y)


def test_sparse_methods():
    # Test sparsity inducing methods. At the moment just checks running.
    latent_dims = 2
    c1 = [1, 3]
    c2 = [1, 3]

    param_candidates = {'c': list(itertools.product(c1, c2))}
    pmd = PMD(latent_dims=latent_dims, random_state=rng).gridsearch_fit(X, Y,
                                                                        param_candidates=param_candidates,
                                                                        verbose=True, plot=True)
    c1 = [1e-4, 1e-5]
    c2 = [1e-4, 1e-5]
    param_candidates = {'c': list(itertools.product(c1, c2))}
    scca = SCCA(latent_dims=latent_dims, random_state=rng).gridsearch_fit(X, Y,
                                                                          param_candidates=param_candidates,
                                                                          verbose=True)
    elastic = ElasticCCA(latent_dims=latent_dims, random_state=rng).gridsearch_fit(X, Y,
                                                                                   param_candidates=param_candidates,
                                                                                   verbose=True)
    corr_pmd = pmd.score(X, Y)
    corr_scca = scca.score(X, Y)
    corr_elastic = elastic.score(X, Y)
    scca_admm = SCCA_ADMM(c=[1e-4, 1e-4]).fit(X, Y)
    scca = SCCA(c=[1e-4, 1e-4]).fit(X, Y)


def test_weighted_GCCA_methods():
    # Test the 'fancy' additions to GCCA i.e. the view weighting and observation weighting.
    latent_dims = 2
    c = 0
    unweighted_gcca = GCCA(latent_dims=latent_dims, c=[c, c]).fit(X, Y)
    deweighted_gcca = GCCA(latent_dims=latent_dims, c=[c, c], view_weights=[0.5, 0.5]).fit(
        X, Y)
    corr_unweighted_gcca = unweighted_gcca.score(X, Y)
    corr_deweighted_gcca = deweighted_gcca.score(X, Y)
    # Check the correlations from each unregularized method are the same
    K = np.ones((2, X.shape[0]))
    K[0, 200:] = 0
    unobserved_gcca = GCCA(latent_dims=latent_dims, c=[c, c]).fit(X, Y, K=K)
    assert np.testing.assert_array_almost_equal(corr_unweighted_gcca, corr_deweighted_gcca, decimal=1) is None


def test_TCCA():
    # Tests tensor CCA methods
    latent_dims = 2
    tcca = TCCA(latent_dims=latent_dims, c=[0.2, 0.2]).fit(X, Y)
    ktcca = KTCCA(latent_dims=latent_dims, c=[0.2, 0.2]).fit(X, Y)
    corr_tcca = tcca.score(X, Y)
    corr_ktcca = ktcca.score(X, Y)
    assert np.testing.assert_array_almost_equal(corr_tcca, corr_ktcca, decimal=1) is None


def test_cv_fit():
    # Test the CV method
    latent_dims = 2
    c1 = [0.1, 0.2]
    c2 = [0.1, 0.2]
    param_candidates = {'c': list(itertools.product(c1, c2))}
    unweighted_gcca = GCCA(latent_dims=latent_dims).gridsearch_fit(X, Y, folds=5,
                                                                   param_candidates=param_candidates,
                                                                   plot=True, jobs=3)
    deweighted_gcca = GCCA(latent_dims=latent_dims, view_weights=[0.5, 0.5]).gridsearch_fit(
        X, Y, folds=2, param_candidates=param_candidates)
    mcca = MCCA(latent_dims=latent_dims).gridsearch_fit(
        X, Y, folds=2, param_candidates=param_candidates)


def test_l0():
    span_cca = SpanCCA(latent_dims=1, regularisation='l0', c=[2, 2]).fit(X, Y)
    swcca = SWCCA(latent_dims=1, c=[2, 2], sample_support=5).fit(X, Y)
    assert (np.abs(span_cca.weights[0]) > 1e-5).sum() == 2
    assert (np.abs(span_cca.weights[1]) > 1e-5).sum() == 2
    assert (np.abs(swcca.weights[0]) > 1e-5).sum() == 2
    assert (np.abs(swcca.weights[1]) > 1e-5).sum() == 2
    assert (np.abs(swcca.loop.sample_weights) > 1e-5).sum() == 5



from sklearn.utils.validation import check_random_state

import cca_zoo.models.innerloop
import cca_zoo.models.innerloop

rng = check_random_state(0)
X = rng.rand(10, 10)
Y = rng.rand(10, 10)
Z = rng.rand(10, 10)


def test_regularized():
    park = cca_zoo.models.innerloop.ParkhomenkoInnerLoop(c=[0.0001, 0.0001]).fit(X, Y)
    park_gen = cca_zoo.models.innerloop.ParkhomenkoInnerLoop(c=[0.0001, 0.0001], generalized=True).fit(X,
                                                                                                       Y)
    params = {'c': [2, 2]}
    pmd = cca_zoo.models.innerloop.PMDInnerLoop(c=[2, 2]).fit(X, Y)
    pmd_gen = cca_zoo.models.innerloop.PMDInnerLoop(c=[2, 2], generalized=True).fit(X, Y)



import numpy as np
from sklearn.utils.validation import check_random_state
from torch import optim, manual_seed
from torch.utils.data import Subset

from cca_zoo import data
from cca_zoo.data import Noisy_MNIST_Dataset
from cca_zoo.deepmodels import DCCA, DCCAE, DVCCA, DCCA_NOI, DTCCA, SplitAE, DeepWrapper
from cca_zoo.deepmodels import objectives, architectures
from cca_zoo.models import CCA

manual_seed(0)
rng = check_random_state(0)
X = rng.rand(200, 10)
Y = rng.rand(200, 10)
Z = rng.rand(200, 10)
X_conv = rng.rand(100, 1, 16, 16)
Y_conv = rng.rand(100, 1, 16, 16)
train_dataset = data.CCA_Dataset(X, Y)


def test_input_types():
    latent_dims = 2
    device = 'cpu'
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    # DCCA
    dcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                      objective=objectives.CCA)
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dcca_model = DeepWrapper(dcca_model, device=device)
    dcca_model.fit(train_dataset, epochs=3)
    dcca_model.fit(train_dataset, val_dataset=train_dataset, epochs=3)
    dcca_model.fit((X, Y), val_dataset=(X, Y), epochs=3)
    dcca_model.fit((X, Y), val_split=0.2, epochs=3)


def tutorial_test():
    # Load MNIST Data
    N = 500
    latent_dims = 2
    dataset = Noisy_MNIST_Dataset(mnist_type='FashionMNIST', train=True)
    ids = np.arange(min(2 * N, len(dataset)))
    np.random.shuffle(ids)
    train_ids, val_ids = np.array_split(ids, 2)
    val_dataset = Subset(dataset, val_ids)
    train_dataset = Subset(dataset, train_ids)
    test_dataset = Noisy_MNIST_Dataset(mnist_type='FashionMNIST', train=False)
    test_ids = np.arange(min(N, len(test_dataset)))
    np.random.shuffle(test_ids)
    test_dataset = Subset(test_dataset, test_ids)
    print('DCCA')
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=784)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=784)
    dcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2])
    dcca_model = DeepWrapper(dcca_model)
    dcca_model.fit(train_dataset, val_dataset=val_dataset, epochs=2)
    dcca_results = np.stack((dcca_model.score(train_dataset), dcca_model.correlations(test_dataset)[0, 1]))


def test_large_p():
    large_p = 256
    X = rng.rand(2000, large_p)
    Y = rng.rand(2000, large_p)
    latent_dims = 32
    device = 'cpu'
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=large_p)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=large_p)
    dcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                      objective=objectives.MCCA, eps=1e-3).float()
    optimizer = optim.Adam(dcca_model.parameters(), lr=1e-4)
    dcca_model = DeepWrapper(dcca_model, device=device, optimizer=optimizer)
    dcca_model.fit((X, Y), epochs=100)


def test_DCCA_methods_cpu():
    latent_dims = 4
    cca_model = CCA(latent_dims=latent_dims).fit(X, Y)
    device = 'cpu'
    epochs = 100
    # DCCA
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    dcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                      objective=objectives.CCA)
    optimizer = optim.SGD(dcca_model.parameters(), lr=1e-2)
    dcca_model = DeepWrapper(dcca_model, device=device, optimizer=optimizer)
    dcca_model.fit((X, Y), epochs=epochs)
    assert np.testing.assert_array_less(cca_model.score(X, Y).sum(),
                                        dcca_model.score((X, Y)).sum()) is None
    # DGCCA
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    dgcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                       objective=objectives.GCCA)
    optimizer = optim.SGD(dgcca_model.parameters(), lr=1e-2)
    dgcca_model = DeepWrapper(dgcca_model, device=device, optimizer=optimizer)
    dgcca_model.fit((X, Y), epochs=epochs)
    assert np.testing.assert_array_less(cca_model.score(X, Y).sum(),
                                        dgcca_model.score((X, Y)).sum()) is None
    # DMCCA
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    dmcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                       objective=objectives.MCCA)
    optimizer = optim.SGD(dmcca_model.parameters(), lr=1e-2)
    dmcca_model = DeepWrapper(dmcca_model, device=device, optimizer=optimizer)
    dmcca_model.fit((X, Y), epochs=epochs)
    assert np.testing.assert_array_less(cca_model.score(X, Y).sum(),
                                        dmcca_model.score((X, Y)).sum()) is None
    # DCCA_NOI
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    dcca_noi_model = DCCA_NOI(latent_dims, X.shape[0], encoders=[encoder_1, encoder_2], rho=0)
    optimizer = optim.Adam(dcca_noi_model.parameters(), lr=1e-2)
    dcca_noi_model = DeepWrapper(dcca_noi_model, device=device, optimizer=optimizer)
    dcca_noi_model.fit((X, Y), epochs=epochs)
    assert np.testing.assert_array_less(cca_model.score(X, Y).sum(),
                                        dcca_noi_model.score((X, Y)).sum()) is None


def test_DTCCA_methods_cpu():
    latent_dims = 2
    device = 'cpu'
    encoder_1 = architectures.Encoder(latent_dims=10, feature_size=10)
    encoder_2 = architectures.Encoder(latent_dims=10, feature_size=10)
    encoder_3 = architectures.Encoder(latent_dims=10, feature_size=10)
    # DTCCA
    dtcca_model = DTCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2])
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dtcca_model = DeepWrapper(dtcca_model, device=device)
    dtcca_model.fit((X, Y), epochs=20)
    # DCCA
    dcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                      objective=objectives.GCCA)
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dcca_model = DeepWrapper(dcca_model, device=device)
    dcca_model.fit((X, Y), epochs=20)


def test_scheduler():
    latent_dims = 2
    device = 'cpu'
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    # DCCA
    dcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                      objective=objectives.CCA)
    optimizer = optim.Adam(dcca_model.parameters(), lr=1e-4)
    scheduler = optim.lr_scheduler.CosineAnnealingLR(optimizer, 1)
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dcca_model = DeepWrapper(dcca_model, device=device, optimizer=optimizer, scheduler=scheduler)
    dcca_model.fit((X, Y), epochs=20)


def test_DGCCA_methods_cpu():
    latent_dims = 2
    device = 'cpu'
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_3 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    # DTCCA
    dtcca_model = DTCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2])
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dtcca_model = DeepWrapper(dtcca_model, device=device)
    dtcca_model.fit((X, Y, Z))
    # DGCCA
    dgcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2, encoder_3],
                       objective=objectives.GCCA)
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dgcca_model = DeepWrapper(dgcca_model, device=device)
    dgcca_model.fit((X, Y, Z))
    # DMCCA
    dmcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2, encoder_3],
                       objective=objectives.MCCA)
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dmcca_model = DeepWrapper(dmcca_model, device=device)
    dmcca_model.fit((X, Y, Z))


def test_DCCAE_methods_cpu():
    latent_dims = 2
    device = 'cpu'
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    decoder_1 = architectures.Decoder(latent_dims=latent_dims, feature_size=10)
    decoder_2 = architectures.Decoder(latent_dims=latent_dims, feature_size=10)
    # DCCAE
    dccae_model = DCCAE(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                        decoders=[decoder_1, decoder_2])
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dccae_model = DeepWrapper(dccae_model, device=device)
    dccae_model.fit((X, Y), epochs=20)
    # SplitAE
    splitae_model = SplitAE(latent_dims=latent_dims, encoder=encoder_1,
                            decoders=[decoder_1, decoder_2])
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    splitae_model = DeepWrapper(splitae_model, device=device)
    splitae_model.fit((X, Y), epochs=10)


def test_DCCAEconv_methods_cpu():
    latent_dims = 2
    device = 'cpu'
    encoder_1 = architectures.CNNEncoder(latent_dims=latent_dims, feature_size=[16, 16])
    encoder_2 = architectures.CNNEncoder(latent_dims=latent_dims, feature_size=[16, 16])
    decoder_1 = architectures.CNNDecoder(latent_dims=latent_dims, feature_size=[16, 16])
    decoder_2 = architectures.CNNDecoder(latent_dims=latent_dims, feature_size=[16, 16])
    # DCCAE
    dccae_model = DCCAE(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                        decoders=[decoder_1, decoder_2])
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dccae_model = DeepWrapper(dccae_model, device=device)
    dccae_model.fit((X_conv, Y_conv))


def test_DVCCA_methods_cpu():
    latent_dims = 2
    device = 'cpu'
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10, variational=True)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10, variational=True)
    decoder_1 = architectures.Decoder(latent_dims=latent_dims, feature_size=10, norm_output=True)
    decoder_2 = architectures.Decoder(latent_dims=latent_dims, feature_size=10, norm_output=True)
    # DVCCA
    dvcca_model = DVCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                        decoders=[decoder_1, decoder_2])
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dvcca_model = DeepWrapper(dvcca_model, device=device)
    dvcca_model.fit((X, Y))


def test_DVCCA_p_methods_cpu():
    latent_dims = 2
    device = 'cpu'
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10, variational=True)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10, variational=True)
    private_encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10, variational=True)
    private_encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10, variational=True)
    decoder_1 = architectures.Decoder(latent_dims=2 * latent_dims, feature_size=10, norm_output=True)
    decoder_2 = architectures.Decoder(latent_dims=2 * latent_dims, feature_size=10, norm_output=True)
    # DVCCA
    dvcca_model = DVCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                        decoders=[decoder_1, decoder_2],
                        private_encoders=[private_encoder_1, private_encoder_2])
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dvcca_model = DeepWrapper(dvcca_model, device=device)
    dvcca_model.fit((X, Y))


def test_DCCA_methods_gpu():
    latent_dims = 2
    device = 'cuda'
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    # DCCA
    dcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                      objective=objectives.CCA)
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dcca_model = DeepWrapper(dcca_model, device=device)
    dcca_model.fit((X, Y))
    # DGCCA
    dgcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                       objective=objectives.GCCA)
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dgcca_model = DeepWrapper(dgcca_model, device=device)
    dgcca_model.fit((X, Y))
    # DMCCA
    dmcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                       objective=objectives.MCCA)
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dmcca_model = DeepWrapper(dmcca_model, device=device)
    dmcca_model.fit((X, Y))
    # DCCA_NOI
    dcca_noi_model = DCCA_NOI(latent_dims, X.shape[0], encoders=[encoder_1, encoder_2], rho=0)
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dcca_noi_model = DeepWrapper(dcca_noi_model, device=device)
    dcca_noi_model.fit((X, Y))


def test_DGCCA_methods_gpu():
    latent_dims = 2
    device = 'cuda'
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_3 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    # DGCCA
    dgcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2, encoder_3],
                       objective=objectives.GCCA)
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dgcca_model = DeepWrapper(dgcca_model, device=device)
    dgcca_model.fit((X, Y, Z))
    # DMCCA
    dmcca_model = DCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2, encoder_3],
                       objective=objectives.MCCA)
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dmcca_model = DeepWrapper(dmcca_model, device=device)
    dmcca_model.fit((X, Y, Z))


def test_DCCAE_methods_gpu():
    latent_dims = 2
    device = 'cuda'
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10)
    decoder_1 = architectures.Decoder(latent_dims=latent_dims, feature_size=10)
    decoder_2 = architectures.Decoder(latent_dims=latent_dims, feature_size=10)
    # DCCAE
    dccae_model = DCCAE(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                        decoders=[decoder_1, decoder_2])
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dccae_model = DeepWrapper(dccae_model, device=device)
    dccae_model.fit((X, Y))


def test_DVCCA_methods_gpu():
    latent_dims = 2
    device = 'cuda'
    encoder_1 = architectures.Encoder(latent_dims=latent_dims, feature_size=10, variational=True)
    encoder_2 = architectures.Encoder(latent_dims=latent_dims, feature_size=10, variational=True)
    decoder_1 = architectures.Decoder(latent_dims=latent_dims, feature_size=10, norm_output=True)
    decoder_2 = architectures.Decoder(latent_dims=latent_dims, feature_size=10, norm_output=True)
    # DVCCA
    dvcca_model = DVCCA(latent_dims=latent_dims, encoders=[encoder_1, encoder_2],
                        decoders=[decoder_1, decoder_2])
    # hidden_layer_sizes are shown explicitly but these are also the defaults
    dvcca_model = DeepWrapper(dvcca_model, device=device)
    dvcca_model.fit((X, Y))



import numpy as np
from sklearn.utils.validation import check_random_state

from cca_zoo.data import Noisy_MNIST_Dataset, Tangled_MNIST_Dataset, Split_MNIST_Dataset
from cca_zoo.data import generate_covariance_data
from cca_zoo.models import MCCA, CCA

rng = check_random_state(0)
X = rng.rand(500, 20)
Y = rng.rand(500, 21)
Z = rng.rand(500, 22)


def test_data_rand():
    (x, y), true_feats = generate_covariance_data(1000, [10, 11], 1, [0.5, 0.5], correlation=0.5,
                                                  structure='random')


def test_data_gen():
    (x, y, z), true_feats = generate_covariance_data(1000, [10, 11, 12], 1, [0.5, 0.5, 0.5], correlation=0.5,
                                                     structure=['identity', 'identity', 'identity'])
    cca = CCA().fit(x[:500], y[:500])
    cca_pred = cca.score(x[500:], y[500:])
    mcca = MCCA().fit(x[:500], y[:500], z[:500])
    mcca_pred = mcca.score(x[500:], y[500:], z[500:])

    (x, y), true_feats = generate_covariance_data(1000, [10, 11], 1, [0.5, 0.5], correlation=0.5,
                                                  structure=['gaussian', 'toeplitz'])


def test_deep_data():
    dataset = Noisy_MNIST_Dataset(mnist_type='FashionMNIST', train=True)
    (train_view_1, train_view_2), (train_rotations, train_labels) = dataset.to_numpy(np.arange(10))
    dataset = Tangled_MNIST_Dataset(mnist_type='FashionMNIST', train=True)
    (train_view_1, train_view_2), (train_rotations_1, train_rotations_2, train_labels) = dataset.to_numpy(
        np.arange(10))
    dataset = Split_MNIST_Dataset(mnist_type='FashionMNIST', train=True)
    (train_view_1, train_view_2), (train_labels) = dataset.to_numpy(np.arange(10))
